@@ -8,11 +8,13 @@
 # US Government Users Restricted Rights - Use, duplication, or
 # disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 ########################################################## {COPYRIGHT-END} ###
+"""Base class for document retrieval for various connections."""
 
 import logging
 import os
 import sys
 import re
+from subprocess import check_call, CalledProcessError
 
 ENCODING = 'utf-8'
 
@@ -20,29 +22,44 @@ ENCODING = 'utf-8'
 class DocumentRetrievalFactory:
     """Factory class to create the right sort of retrieval object."""
 
+    @staticmethod
+    def _get_matching_connection(datasource, cluster, connections):
+        """Get connection matching datasource and cluster."""
+        for conn in connections:
+            if conn['datasource'] == datasource and conn['cluster'] == cluster:
+                return conn
+        return None
+
     @staticmethod
     def create(application, key):
         """Lookup connection and create required type."""
-        platform, client = application.connections.get((key.datasource, key.cluster), (None, None))
-
-        if platform and client:
+        connection = DocumentRetrievalFactory._get_matching_connection(key.datasource, key.cluster, application.conn_details)
+        try:
+            platform = connection['platform']
             if platform == 'COS':
-                return DocumentRetrievalCOS(client)
+                _, client, connection = application.create_cos_connection(connection)
+                return DocumentRetrievalCOS(client, connection)
             elif platform == 'NFS':
-                return DocumentRetrievalNFS(client)
+                _, client, connection = application.create_nfs_connection(connection)
+                return DocumentRetrievalNFS(client, connection)
             elif platform == 'Spectrum Scale':
-                return DocumentRetrievalScale(client)
-            elif platform == 'Spectrum Scale Local':
-                return DocumentRetrievalLocalScale(client)
+                platform, client, connection = application.create_scale_connection(connection)
+                if platform == 'Spectrum Scale Local':
+                    return DocumentRetrievalLocalScale(client, connection)
+                return DocumentRetrievalScale(client, connection)
             elif platform == 'SMB/CIFS':
-                return DocumentRetrievalSMB(client)
+                _, client, connection = application.create_smb_connection(connection, key.fileset)
+                return DocumentRetrievalSMB(client, connection)
+        except KeyError:
+            # No connection found
+            pass
         return None
 
 
 class DocumentRetrievalBase():
     """A class to retrieve a document via a Spectrum Discover Connection."""
 
-    def __init__(self, client):
+    def __init__(self, client, connection):
         """Constructor."""
         # Instantiate logger
         loglevels = {'INFO': logging.INFO, 'DEBUG': logging.DEBUG,
@@ -55,6 +72,15 @@ class DocumentRetrievalBase():
         self.logger = logging.getLogger(__name__)
 
         self.client = client
+        self.connection = connection
+
+        self.stat_atime = None
+        self.stat_mtime = None
+
+        try:
+            self.preserve_stat_time = self.connection['additional_info']['preserve_stat_time']
+        except (KeyError, TypeError):
+            self.preserve_stat_time = False
 
     def get_document(self, key):
         """
@@ -67,7 +93,6 @@ class DocumentRetrievalBase():
         Returns None if object was not able to be retrieved.
         """
         self.logger.warning("get_document has not been implemented for this class")
-        return None
 
     def get_headers(self, key):
         """
@@ -79,7 +104,6 @@ class DocumentRetrievalBase():
         Returns None if object was not able to be retrieved.
         """
         self.logger.warning("get_headers has not been implemented for this class")
-        return None
 
     def cleanup_document(self):
         """
@@ -91,6 +115,15 @@ class DocumentRetrievalBase():
         """
         self.logger.warning("cleanup_document has not been implemented for this class")
 
+    def close_connection(self):
+        """
+        To be implemented by the child class if needed.
+
+        Closes an open connection.
+
+        Returns None.
+        """
+
     def create_file_path(self, prefix, filetype):
         """Return the tmpfile_path with filetype as string."""
         self.logger.debug('filepath prefix: %s, type: %s', prefix, filetype)
@@ -98,6 +131,44 @@ class DocumentRetrievalBase():
             return prefix + "." + filetype
         return prefix
 
+    def save_stat_times(self, filepath, method=os, nanoseconds=True):
+        """
+        Save the stat info for the given file.
+
+        If nanoseconds=True, nano-second results will be preserved, else seconds will be preserved.
+        By default, use the os module. For remote scale connections, use paramiko sftp module.
+        """
+        self.logger.debug('Attempting to get stat info for %s', filepath)
+        try:
+            stat = method.stat(filepath)
+            if nanoseconds:
+                self.stat_atime = stat.st_atime_ns
+                self.stat_mtime = stat.st_mtime_ns
+            else:
+                self.stat_atime = stat.st_atime
+                self.stat_mtime = stat.st_mtime
+
+            self.logger.debug('Successfully retrieved stat info for %s. atime: %s, mtime: %s', filepath, str(self.stat_atime),
+                              str(self.stat_mtime))
+        except (PermissionError, FileNotFoundError) as error:
+            self.logger.error('Failed to retrieve stat info for %s. Error: %s', filepath, str(error))
+            self.stat_atime = None
+            self.stat_mtime = None
+
+    def restore_stat_times(self, filepath, method=os, nanoseconds=True):
+        """Restore the stat info (atime, mtime) for the given file."""
+        if self.stat_atime and self.stat_mtime:
+            self.logger.debug('Attempting to restore stat info for %s', filepath)
+            try:
+                if nanoseconds:
+                    method.utime(filepath, ns=(self.stat_atime, self.stat_mtime))
+                else:
+                    method.utime(filepath, (self.stat_atime, self.stat_mtime))
+
+                self.logger.debug('Successfully restored stat info for %s. atime: %s, mtime: %s', filepath, str(self.stat_atime),
+                                  str(self.stat_mtime))
+            except (PermissionError, FileNotFoundError, OSError) as error:
+                self.logger.error('Failed to restore stat info for %s. Error: %s', filepath, str(error))
 
 class DocumentRetrievalCOS(DocumentRetrievalBase):
     """Create a COS document class.
@@ -161,19 +232,35 @@ class DocumentRetrievalNFS(DocumentRetrievalBase):
         """Return document filepath."""
         self.filepath = None
 
-        if self.client:
-            mount_path_prefix = self.client['additional_info']['local_mount']
-            source_path_prefix = self.client['mount_point']
+        if self.connection:
+            mount_path_prefix = self.connection['additional_info']['local_mount']
+            source_path_prefix = self.connection['mount_point']
             self.filepath = re.sub('^' + source_path_prefix, mount_path_prefix, key.path.decode(ENCODING))
         else:
             self.logger.info('No document match')
 
+        if self.preserve_stat_time:
+            self.save_stat_times(self.filepath)
+
         return self.filepath
 
     def cleanup_document(self):
         """Cleanup files as needed."""
+        if self.preserve_stat_time:
+            self.restore_stat_times(self.filepath)
+
         self.logger.debug("NFS: Not doing any cleanup.")
 
+    def close_connection(self):
+        """Unmounts an nfs connection."""
+        mount_point = self.connection['additional_info']['local_mount']
+        cmd = f'umount {mount_point}'
+        try:
+            check_call(cmd, shell=True)
+        except CalledProcessError as cpe:
+            self.logger.warning('Failed to to unmount NFS export %s for connection %s. Error: %s', mount_point,
+                                self.connection['name'], str(cpe))
+
 
 class DocumentRetrievalScale(DocumentRetrievalBase):
     """Create a Spectrum Scale document class.
@@ -182,22 +269,26 @@ class DocumentRetrievalScale(DocumentRetrievalBase):
     """
 
     filepath = None
+    key = None
 
     def get_document(self, key):
         """Return document filepath."""
         self.filepath = None
+        self.key = key
+
+        if self.preserve_stat_time:
+            self.save_stat_times(self.key.path.decode(ENCODING), method=self.client, nanoseconds=False)
 
         if self.client:
             try:
-                self.filepath = self.create_file_path('/tmp/scalefile_' + str(os.getpid()), key.filetype)
-                self.logger.debug(self.filepath)
-                self.client.get(key.path, self.filepath)
+                self.filepath = self.create_file_path('/tmp/scalefile_' + str(os.getpid()), self.key.filetype)
+                self.client.get(self.key.path, self.filepath)
             except UnicodeDecodeError:
-                self.logger.error('Could not decode file %s', key.path.decode(ENCODING))
+                self.logger.error('Could not decode file %s', self.key.path.decode(ENCODING))
             except FileNotFoundError:
-                self.logger.error('Could not find file %s', key.path.decode(ENCODING))
+                self.logger.error('Could not find file %s', self.key.path.decode(ENCODING))
             except OSError:  # Seen when scale nsd disk is down and file is not accessible
-                self.logger.error('Could not transfer file %s', key.path.decode(ENCODING))
+                self.logger.error('Could not transfer file %s', self.key.path.decode(ENCODING))
 
         else:
             self.logger.info('No document match')
@@ -208,11 +299,18 @@ class DocumentRetrievalScale(DocumentRetrievalBase):
         """Cleanup files as needed."""
         self.logger.debug("SCALE: Attempting to delete file: %s", self.filepath)
 
+        if self.preserve_stat_time:
+            self.restore_stat_times(self.key.path.decode(ENCODING), method=self.client, nanoseconds=False)
+
         try:
             os.remove(self.filepath)
         except FileNotFoundError:
             self.logger.debug('No file: %s to delete.', self.filepath)
 
+    def close_connection(self):
+        """Close an active sftp connection."""
+        self.client.close()
+
 
 class DocumentRetrievalLocalScale(DocumentRetrievalBase):
     """Create a Local Spectrum Scale document class.
@@ -221,12 +319,22 @@ class DocumentRetrievalLocalScale(DocumentRetrievalBase):
     No need to download or cleanup since accessed direclty upon the mount point.
     """
 
+    filepath = None
+
     def get_document(self, key):
         """Return Document Key."""
-        return key.path.decode(ENCODING)
+        self.filepath = key.path.decode(ENCODING)
+
+        if self.preserve_stat_time:
+            self.save_stat_times(self.filepath)
+
+        return self.filepath
 
     def cleanup_document(self):
         """Cleanup files as needed."""
+        if self.preserve_stat_time:
+            self.restore_stat_times(self.filepath)
+
         self.logger.debug("Scale Local: Not doing any cleanup.")
 
 
@@ -243,18 +351,34 @@ class DocumentRetrievalSMB(DocumentRetrievalBase):
         """Return document filepath."""
         self.filepath = None
 
-        if self.client:
-            mount_path_prefix = self.client['additional_info']['local_mount']
+        if self.connection:
+            mount_path_prefix = self.connection['additional_info']['local_mount']
             self.filepath = mount_path_prefix + key.path.decode(ENCODING)
         else:
             self.logger.info('No document match')
 
+        if self.preserve_stat_time:
+            self.save_stat_times(self.filepath)
+
         return self.filepath
 
     def cleanup_document(self):
         """Cleanup files as needed."""
+        if self.preserve_stat_time:
+            self.restore_stat_times(self.filepath)
+
         self.logger.debug("SMB: Not doing any cleanup.")
 
+    def close_connection(self):
+        """Unmounts a smb connection."""
+        mount_point = self.connection['additional_info']['local_mount']
+        cmd = f'umount {mount_point}'
+        try:
+            check_call(cmd, shell=True)
+        except CalledProcessError as cpe:
+            self.logger.warning('Failed to to unmount SMB export %s for connection %s. Error: %s', mount_point,
+                                self.connection['name'], str(cpe))
+
 
 class DocumentKey(object):
     """A class to identify a unique document on a Spectrum Discover Connection."""
@@ -265,7 +389,10 @@ class DocumentKey(object):
         self.datasource = doc['datasource']
         self.cluster = doc['cluster']
         self.path = doc['path'].encode(ENCODING)
-        self.filetype = doc['type']
+        if 'type' in doc.keys():  # deepinspect
+            self.filetype = doc['type']
+        if 'fileset' in doc.keys(): # 2.0.3+
+            self.fileset = doc['fileset']
         # a unique identifier for the connection this document belongs to.
         self.id = self.datasource + ':' + self.cluster
 
